@@ -6,7 +6,7 @@ import os
 import tempfile
 import threading
 import time
-from collections import OrderedDict
+from collections import OrderedDict, defaultdict, deque
 from contextlib import ExitStack
 from functools import wraps
 from typing import Union, Tuple, List, Set, Dict, Iterator, Callable, Type, TextIO, Any
@@ -16,8 +16,8 @@ from ruamel.yaml import StringIO
 
 from .. import JINA_GLOBAL
 from ..enums import FlowBuildLevel, FlowOptimizeLevel
-from ..excepts import FlowTopologyError, FlowMissingPodError, FlowBuildLevelError, FlowConnectivityError
-from ..helper import yaml, expand_env_var, get_non_defaults_args, deprecated_alias
+from ..excepts import FlowTopologyError, FlowMissingPodError, FlowBuildLevelError
+from ..helper import yaml, expand_env_var, get_non_defaults_args, deprecated_alias, random_port
 from ..logging import get_logger
 from ..logging.sse import start_sse_logger
 from ..peapods.pod import SocketType, FlowPod, GatewayFlowPod
@@ -62,6 +62,87 @@ def build_required(required_level: 'FlowBuildLevel'):
     return __build_level
 
 
+def _traverse_graph(op_flow: 'Flow', outgoing_map: Dict[str, List[str]],
+                    func: Callable[['Flow', str, str], None]) -> 'Flow':
+    _outgoing_idx = dict.fromkeys(outgoing_map.keys(), 0)
+    stack = deque()
+    stack.append('gateway')
+    op_flow.logger.debug('Traversing dependency graph:')
+    while stack:
+        start_node_name = stack.pop()
+        end_node_idx = _outgoing_idx[start_node_name]
+        if end_node_idx < len(outgoing_map[start_node_name]):
+            # else, you are back to the gateway
+            end_node_name = outgoing_map[start_node_name][end_node_idx]
+            func(op_flow, start_node_name, end_node_name)
+            stack.append(end_node_name)
+            if end_node_idx + 1 < len(outgoing_map[start_node_name]):
+                stack.append(start_node_name)
+            _outgoing_idx[start_node_name] = end_node_idx + 1
+    return op_flow
+
+
+def _build_flow(op_flow: 'Flow', outgoing_map: Dict[str, List[str]]) -> 'Flow':
+    def _build_two_connections(flow: 'Flow', start_node_name: str, end_node_name: str):
+        # Rule
+        # if a node has multiple income/outgoing peas,
+        # then its socket_in/out must be PULL_BIND or PUB_BIND
+        # otherwise it should be different than its income
+        # i.e. income=BIND => this=CONNECT, income=CONNECT => this = BIND
+        #
+        # when a socket is BIND, then host must NOT be set, aka default host 0.0.0.0
+        # host_in and host_out is only set when corresponding socket is CONNECT
+        start_node = flow._pod_nodes[start_node_name]
+        end_node = flow._pod_nodes[end_node_name]
+        first_socket_type = SocketType.PUSH_CONNECT
+        if len(outgoing_map[start_node_name]) > 1:
+            first_socket_type = SocketType.PUB_BIND
+        elif end_node_name == 'gateway':
+            first_socket_type = SocketType.PUSH_BIND
+        flow.logger.debug(f'Connect {start_node_name} with {end_node_name}')
+        FlowPod.connect(start_node, end_node, first_socket_type=first_socket_type)
+
+    return _traverse_graph(op_flow, outgoing_map, _build_two_connections)
+
+
+def _optimize_flow(op_flow, outgoing_map: Dict[str, List[str]], pod_edges: {str, str}) -> 'Flow':
+    def _optimize_two_connections(flow: 'Flow', start_node_name: str, end_node_name: str):
+        start_node = flow._pod_nodes[start_node_name]
+        end_node = flow._pod_nodes[end_node_name]
+        edges_with_same_start = [ed for ed in pod_edges if ed[0].startswith(start_node_name)]
+        edges_with_same_end = [ed for ed in pod_edges if ed[1].endswith(end_node_name)]
+        if len(edges_with_same_start) > 1 or len(edges_with_same_end) > 1:
+            flow.logger.info(f'Connection between {start_node_name} and {end_node_name} cannot be optimized')
+        else:
+            if start_node_name == 'gateway':
+                if flow.args.optimize_level > FlowOptimizeLevel.IGNORE_GATEWAY and end_node.is_head_router:
+                    flow.logger.info(
+                        f'Node {end_node_name} connects to tail of {start_node_name}')
+                    end_node.connect_to_tail_of(start_node)
+            elif end_node_name == 'gateway':
+                if flow.args.optimize_level > FlowOptimizeLevel.IGNORE_GATEWAY and \
+                        start_node.is_tail_router and start_node.tail_args.num_part <= 1:
+                    # connect gateway directly to peas only if this is unblock router
+                    # as gateway can not block & reduce message
+                    flow.logger.info(
+                        f'Node {start_node_name} connects to head of {end_node_name}')
+                    start_node.connect_to_head_of(end_node)
+            else:
+                if end_node.is_head_router and not start_node.is_tail_router:
+                    flow.logger.info(
+                        f'Node {end_node_name} connects to tail of {start_node_name}')
+                    end_node.connect_to_tail_of(start_node)
+                elif start_node.is_tail_router and start_node.tail_args.num_part <= 1:
+                    flow.logger.info(
+                        f'Node {start_node_name} connects to head of {end_node_name}')
+                    start_node.connect_to_head_of(end_node)
+
+    if op_flow.args.optimize_level > FlowOptimizeLevel.NONE:
+        return _traverse_graph(op_flow, outgoing_map, _optimize_two_connections)
+    else:
+        return op_flow
+
+
 class Flow:
     def __init__(self, args: 'argparse.Namespace' = None, **kwargs):
         """Initialize a flow object
@@ -76,7 +157,7 @@ class Flow:
         .. highlight:: python
         .. code-block:: python
 
-            f = Flow(optimize_level=FlowOptimizeLevel.NONE).add(yaml_path='forward', replicas=3)
+            f = Flow(optimize_level=FlowOptimizeLevel.NONE).add(uses='forward', parallel=3)
 
         The optimized version, i.e. :code:`Flow(optimize_level=FlowOptimizeLevel.FULL)`
         will generate 4 Peas, but it will force the :class:`GatewayPea` to take BIND role,
@@ -90,6 +171,7 @@ class Flow:
         self._last_changed_pod = ['gateway']  #: default first pod is gateway, will add when build()
 
         self._update_args(args, **kwargs)
+        self._ports_in_use = []
 
     def _update_args(self, args, **kwargs):
         from ..main.parser import set_flow_parser
@@ -265,18 +347,21 @@ class Flow:
         kwargs['name'] = 'gateway'
         self._pod_nodes[pod_name] = GatewayFlowPod(kwargs, needs)
 
-        # self.set_last_pod(pod_name, False)
-
-    def join(self, needs: Union[Tuple[str], List[str]], *args, **kwargs) -> 'Flow':
+    def needs(self, needs: Union[Tuple[str], List[str]], uses: str = '_merge', name: str = 'joiner', *args,
+              copy_flow: bool = True, **kwargs) -> 'Flow':
         """
         Add a blocker to the flow, wait until all peas defined in **needs** completed.
 
         :param needs: list of service names to wait
+        :param uses: the config of the executor, by default is ``_merge``
+        :param name: the name of this joiner, by default is ``joiner``
         :return: the modified flow
         """
+        op_flow = copy.deepcopy(self) if copy_flow else self
+
         if len(needs) <= 1:
             raise FlowTopologyError('no need to wait for a single service, need len(needs) > 1')
-        return self.add(name='joiner', yaml_path='_merge', needs=needs, *args, **kwargs)
+        return op_flow.add(name=name, uses=uses, needs=needs, *args, **kwargs)
 
     def add(self,
             needs: Union[str, Tuple[str], List[str]] = None,
@@ -316,8 +401,8 @@ class Flow:
 
         kwargs.update(op_flow._common_kwargs)
         kwargs['name'] = pod_name
+        op_flow._ports_in_use += op_flow._check_port_collision(kwargs)
         op_flow._pod_nodes[pod_name] = FlowPod(kwargs=kwargs, needs=needs)
-
         op_flow.set_last_pod(pod_name, False)
 
         return op_flow
@@ -356,66 +441,17 @@ class Flow:
         if 'gateway' not in op_flow._pod_nodes:
             op_flow._add_gateway(needs={op_flow._last_changed_pod[-1]})
 
-        # direct all income peas' output to the current service
-        for k, p in op_flow._pod_nodes.items():
-            for s in p.needs:
-                if s not in op_flow._pod_nodes:
-                    raise FlowMissingPodError(f'{s} is not in this flow, misspelled name?')
-                _pod_edges.add(f'{s}-{k}')
-
-        for k in _pod_edges:
-            s_name, e_name = k.split('-')
-            edges_with_same_start = [ed for ed in _pod_edges if ed.startswith(s_name)]
-            edges_with_same_end = [ed for ed in _pod_edges if ed.endswith(e_name)]
-
-            s_pod = op_flow._pod_nodes[s_name]
-            e_pod = op_flow._pod_nodes[e_name]
-
-            # Rule
-            # if a node has multiple income/outgoing peas,
-            # then its socket_in/out must be PULL_BIND or PUB_BIND
-            # otherwise it should be different than its income
-            # i.e. income=BIND => this=CONNECT, income=CONNECT => this = BIND
-            #
-            # when a socket is BIND, then host must NOT be set, aka default host 0.0.0.0
-            # host_in and host_out is only set when corresponding socket is CONNECT
-
-            if len(edges_with_same_start) > 1 and len(edges_with_same_end) == 1:
-                FlowPod.connect(s_pod, e_pod, first_socket_type=SocketType.PUB_BIND)
-            elif len(edges_with_same_start) == 1 and len(edges_with_same_end) > 1:
-                FlowPod.connect(s_pod, e_pod, first_socket_type=SocketType.PUSH_CONNECT)
-            elif len(edges_with_same_start) == 1 and len(edges_with_same_end) == 1:
-                # in this case, either side can be BIND
-                # we prefer gateway to be always CONNECT so that multiple clients can connect to it
-                # check if either node is gateway
-                # this is the only place where gateway appears
-                if s_name == 'gateway':
-                    if self.args.optimize_level > FlowOptimizeLevel.IGNORE_GATEWAY and e_pod.is_head_router:
-                        # connect gateway directly to peas
-                        e_pod.connect_to_tail_of(s_pod)
-                    else:
-                        FlowPod.connect(s_pod, e_pod, first_socket_type=SocketType.PUSH_CONNECT)
-                elif e_name == 'gateway':
-                    if self.args.optimize_level > FlowOptimizeLevel.IGNORE_GATEWAY and s_pod.is_tail_router and s_pod.tail_args.num_part <= 1:
-                        # connect gateway directly to peas only if this is unblock router
-                        # as gateway can not block & reduce message
-                        s_pod.connect_to_head_of(e_pod)
-                    else:
-                        FlowPod.connect(s_pod, e_pod, first_socket_type=SocketType.PUSH_BIND)
-                else:
-                    e_pod.head_args.socket_in = s_pod.tail_args.socket_out.paired
-                    if self.args.optimize_level > FlowOptimizeLevel.NONE and e_pod.is_head_router and not s_pod.is_tail_router:
-                        e_pod.connect_to_tail_of(s_pod)
-                    elif self.args.optimize_level > FlowOptimizeLevel.NONE and s_pod.is_tail_router and s_pod.tail_args.num_part <= 1:
-                        s_pod.connect_to_head_of(e_pod)
-                    else:
-                        FlowPod.connect(s_pod, e_pod, first_socket_type=SocketType.PUSH_CONNECT)
-            else:
-                raise FlowTopologyError('found %d edges start with %s and %d edges end with %s, '
-                                        'this type of topology is ambiguous and should not exist, '
-                                        'i can not determine the socket type' % (
-                                            len(edges_with_same_start), s_name, len(edges_with_same_end), e_name))
+        # construct a map with a key a start node and values an array of its end nodes
+        _outgoing_map = defaultdict(list)
+        for end, pod in op_flow._pod_nodes.items():
+            for start in pod.needs:
+                if start not in op_flow._pod_nodes:
+                    raise FlowMissingPodError(f'{start} is not in this flow, misspelled name?')
+                _outgoing_map[start].append(end)
+                _pod_edges.add((start, end))
 
+        op_flow = _build_flow(op_flow, _outgoing_map)
+        op_flow = _optimize_flow(op_flow, _outgoing_map, _pod_edges)
         op_flow._build_level = FlowBuildLevel.GRAPH
         return op_flow
 
@@ -482,17 +518,14 @@ class Flow:
 
     @property
     def num_peas(self) -> int:
-        """Get the number of peas (replicas count) in this flow"""
+        """Get the number of peas (parallel count) in this flow"""
         return sum(v.num_peas for v in self._pod_nodes.values())
 
     def close(self):
         """Close the flow and release all resources associated to it. """
         if hasattr(self, '_pod_stack'):
             self._pod_stack.close()
-        # if hasattr(self, 'sse_logger') and self.sse_logger.is_alive():
-        #     self.sse_logger.stop()
         self._build_level = FlowBuildLevel.EMPTY
-        # time.sleep(1)  # sleep for a while until all resources are safely closed
         self.logger.success(
             f'flow is closed and all resources should be released already, current build level is {self._build_level}')
 
@@ -564,11 +597,11 @@ class Flow:
         :param output_fn: the callback function to invoke after training
         :param kwargs: accepts all keyword arguments of `jina client` CLI
         """
-        self._get_client(**kwargs).train(input_fn, output_fn)
+        self._get_client(**kwargs).train(input_fn, output_fn, **kwargs)
 
     def index_ndarray(self, array: 'np.ndarray', axis: int = 0, size: int = None, shuffle: bool = False,
-                    output_fn: Callable[['jina_pb2.Message'], None] = None,
-                    **kwargs):
+                      output_fn: Callable[['jina_pb2.Message'], None] = None,
+                      **kwargs):
         """Using numpy ndarray as the index source for the current flow
 
         :param array: the numpy ndarray data source
@@ -579,11 +612,11 @@ class Flow:
         :param kwargs: accepts all keyword arguments of `jina client` CLI
         """
         from ..clients.python.io import input_numpy
-        self._get_client(**kwargs).index(input_numpy(array, axis, size, shuffle), output_fn)
+        self._get_client(**kwargs).index(input_numpy(array, axis, size, shuffle), output_fn, **kwargs)
 
-    def search_numpy(self, array: 'np.ndarray', axis: int = 0, size: int = None, shuffle: bool = False,
-                     output_fn: Callable[['jina_pb2.Message'], None] = None,
-                     **kwargs):
+    def search_ndarray(self, array: 'np.ndarray', axis: int = 0, size: int = None, shuffle: bool = False,
+                       output_fn: Callable[['jina_pb2.Message'], None] = None,
+                       **kwargs):
         """Use a numpy ndarray as the query source for searching on the current flow
 
         :param array: the numpy ndarray data source
@@ -594,7 +627,7 @@ class Flow:
         :param kwargs: accepts all keyword arguments of `jina client` CLI
         """
         from ..clients.python.io import input_numpy
-        self._get_client(**kwargs).search(input_numpy(array, axis, size, shuffle), output_fn)
+        self._get_client(**kwargs).search(input_numpy(array, axis, size, shuffle), output_fn, **kwargs)
 
     def index_lines(self, lines: Iterator[str] = None, filepath: str = None, size: int = None,
                     sampling_rate: float = None, read_mode='r',
@@ -612,7 +645,8 @@ class Flow:
         :param kwargs: accepts all keyword arguments of `jina client` CLI
         """
         from ..clients.python.io import input_lines
-        self._get_client(**kwargs).index(input_lines(lines, filepath,  size, sampling_rate, read_mode), output_fn)
+        self._get_client(**kwargs).index(input_lines(lines, filepath, size, sampling_rate, read_mode), output_fn,
+                                         **kwargs)
 
     def index_files(self, patterns: Union[str, List[str]], recursive: bool = True,
                     size: int = None, sampling_rate: float = None, read_mode: str = None,
@@ -626,12 +660,13 @@ class Flow:
         :param size: the maximum number of the files
         :param sampling_rate: the sampling rate between [0, 1]
         :param read_mode: specifies the mode in which the file
-                is opened. 'r' for reading in text mode, 'rb' for reading in
+                is opened. 'r' for reading in text mode, 'rb' for reading in binary mode
         :param output_fn: the callback function to invoke after indexing
         :param kwargs: accepts all keyword arguments of `jina client` CLI
         """
         from ..clients.python.io import input_files
-        self._get_client(**kwargs).index(input_files(patterns, recursive, size, sampling_rate, read_mode), output_fn)
+        self._get_client(**kwargs).index(input_files(patterns, recursive, size, sampling_rate, read_mode), output_fn,
+                                         **kwargs)
 
     def search_files(self, patterns: Union[str, List[str]], recursive: bool = True,
                      size: int = None, sampling_rate: float = None, read_mode: str = None,
@@ -650,7 +685,8 @@ class Flow:
         :param kwargs: accepts all keyword arguments of `jina client` CLI
         """
         from ..clients.python.io import input_files
-        self._get_client(**kwargs).search(input_files(patterns, recursive, size, sampling_rate, read_mode), output_fn)
+        self._get_client(**kwargs).search(input_files(patterns, recursive, size, sampling_rate, read_mode), output_fn,
+                                          **kwargs)
 
     def search_lines(self, filepath: str = None, lines: Iterator[str] = None, size: int = None,
                      sampling_rate: float = None, read_mode='r',
@@ -668,7 +704,8 @@ class Flow:
         :param kwargs: accepts all keyword arguments of `jina client` CLI
         """
         from ..clients.python.io import input_lines
-        self._get_client(**kwargs).search(input_lines(lines, filepath, size, sampling_rate, read_mode), output_fn)
+        self._get_client(**kwargs).search(input_lines(lines, filepath, size, sampling_rate, read_mode), output_fn,
+                                          **kwargs)
 
     @deprecated_alias(buffer='input_fn', callback='output_fn')
     def index(self, input_fn: Union[Iterator['jina_pb2.Document'], Iterator[bytes], Callable] = None,
@@ -708,7 +745,7 @@ class Flow:
         :param output_fn: the callback function to invoke after indexing
         :param kwargs: accepts all keyword arguments of `jina client` CLI
         """
-        self._get_client(**kwargs).index(input_fn, output_fn)
+        self._get_client(**kwargs).index(input_fn, output_fn, **kwargs)
 
     @deprecated_alias(buffer='input_fn', callback='output_fn')
     def search(self, input_fn: Union[Iterator['jina_pb2.Document'], Iterator[bytes], Callable] = None,
@@ -749,13 +786,13 @@ class Flow:
         :param output_fn: the callback function to invoke after searching
         :param kwargs: accepts all keyword arguments of `jina client` CLI
         """
-        self._get_client(**kwargs).search(input_fn, output_fn)
+        self._get_client(**kwargs).search(input_fn, output_fn, **kwargs)
 
     def dry_run(self, **kwargs):
         """Send a DRYRUN request to this flow, passing through all pods in this flow,
         useful for testing connectivity and debugging"""
-        if not self._get_client(**kwargs).dry_run():
-            raise FlowConnectivityError('a dry run shows this flow is badly connected due to the network settings')
+        self.logger.warning('calling dry_run() on a flow is depreciated, it will be removed in the future version. '
+                            'calling index(), search(), train() will trigger a dry_run()')
 
     @build_required(FlowBuildLevel.GRAPH)
     def to_swarm_yaml(self, path: TextIO):
@@ -770,7 +807,7 @@ class Flow:
         for k, v in self._pod_nodes.items():
             swarm_yml['services'][k] = {
                 'command': v.to_cli_command(),
-                'deploy': {'replicas': 1}
+                'deploy': {'parallel': 1}
             }
 
         yaml.dump(swarm_yml, path)
@@ -804,3 +841,21 @@ class Flow:
     def use_rest_gateway(self):
         """Change to use REST gateway for IO """
         self._common_kwargs['rest_api'] = True
+
+    def _check_port_collision(self, kwargs):
+        """Check if the Pods' ports collide"""
+        for _port_name in ('port_in', 'port_out', 'port_ctrl', 'port_expose'):
+            _port = kwargs.get(_port_name)
+            if _port is None:
+                _port = random_port()
+            while _port in self._ports_in_use:
+                _new = random_port()
+                _port = _new
+                self.logger.warning(f'{_port_name} collision detected. set from {_port} to {_new}')
+            kwargs[_port_name] = _port
+            self._ports_in_use.append(_port)
+            self.logger.debug(f'{kwargs.get("name")} {_port_name}: {_port}')
+        return self._ports_in_use
+
+    # for backward support
+    join = needs
