@@ -7,15 +7,17 @@ import time
 from contextlib import ExitStack
 from queue import Empty
 from threading import Thread
-from typing import Set, Dict, List, Callable, Union
+from typing import Optional, Set, Dict, List, Callable, Union
 
 from . import Pea
 from .gateway import GatewayPea, RESTGatewayPea
-from .pea import BasePea
+from .head_pea import HeadPea
+from .tail_pea import TailPea
 from .. import __default_host__
 from ..enums import *
-from ..helper import random_port, get_random_identity, get_parsed_args, get_non_defaults_args
+from ..helper import random_port, get_random_identity, get_parsed_args, get_non_defaults_args, valid_local_config_source
 from ..main.parser import set_pod_parser, set_gateway_parser
+from argparse import Namespace
 
 
 class BasePod:
@@ -35,7 +37,12 @@ class BasePod:
         self.deducted_tail = None
         if hasattr(args, 'polling') and args.polling.is_push:
             # ONLY reset when it is push
-            args.reducing_yaml_path = '_forward'
+            args.uses_reducing = '_pass'
+
+        if getattr(args, 'parallel', 1) > 1:
+            self.is_head_router = True
+            self.is_tail_router = True
+
         self._args = args
         self.peas_args = self._parse_args(args)
 
@@ -68,21 +75,19 @@ class BasePod:
         """Get the grpc host name """
         return self.peas_args['peas'][0].host
 
-    def _parse_args(self, args):
+    def _parse_args(self, args: Namespace) -> Dict[str, Optional[Union[List[Namespace], Namespace]]]:
         peas_args = {
             'head': None,
             'tail': None,
             'peas': []
         }
 
-        if getattr(args, 'replicas', 1) > 1:
+        if getattr(args, 'parallel', 1) > 1:
             # reasons to separate head and tail from peas is that they
             # can be deducted based on the previous and next pods
             peas_args['head'] = _copy_to_head_args(args, args.polling.is_push)
             peas_args['tail'] = _copy_to_tail_args(args)
             peas_args['peas'] = _set_peas_args(args, peas_args['head'], peas_args['tail'])
-            self.is_head_router = True
-            self.is_tail_router = True
         else:
             peas_args['peas'] = [args]
 
@@ -138,7 +143,7 @@ class BasePod:
             raise ValueError('ambiguous tail node, maybe it is deducted already?')
 
     @property
-    def all_args(self):
+    def all_args(self) -> List[Namespace]:
         """Get all arguments of all Peas in this BasePod. """
         return self.peas_args['peas'] + (
             [self.peas_args['head']] if self.peas_args['head'] else []) + (
@@ -163,7 +168,7 @@ class BasePod:
             # s.host_in = __default_host__
             # s.host_out = __default_host__
 
-    def start_sentinels(self):
+    def start_sentinels(self) -> None:
         self.sentinel_threads = []
         if isinstance(self._args, argparse.Namespace) and getattr(self._args, 'shutdown_idle', False):
             self.sentinel_threads.append(Thread(target=self.close_if_idle,
@@ -172,7 +177,7 @@ class BasePod:
         for t in self.sentinel_threads:
             t.start()
 
-    def start(self):
+    def start(self) -> 'FlowPod':
         """Start to run all Peas in this BasePod.
 
         Remember to close the BasePod with :meth:`close`.
@@ -183,12 +188,12 @@ class BasePod:
         self.stack = ExitStack()
         # start head and tail
         if self.peas_args['head']:
-            p = BasePea(self.peas_args['head'])
+            p = HeadPea(self.peas_args['head'])
             self.peas.append(p)
             self.stack.enter_context(p)
 
         if self.peas_args['tail']:
-            p = BasePea(self.peas_args['tail'])
+            p = TailPea(self.peas_args['tail'])
             self.peas.append(p)
             self.stack.enter_context(p)
 
@@ -233,7 +238,7 @@ class BasePod:
     def is_shutdown(self) -> bool:
         return all(not p.is_ready.is_set() for p in self.peas)
 
-    def __enter__(self):
+    def __enter__(self) -> Union['GatewayFlowPod', 'FlowPod']:
         return self.start()
 
     @property
@@ -250,7 +255,7 @@ class BasePod:
             p.is_ready.wait()
         return True
 
-    def __exit__(self, exc_type, exc_val, exc_tb):
+    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
         self.close()
 
     def join(self):
@@ -263,7 +268,7 @@ class BasePod:
         finally:
             self.peas.clear()
 
-    def close(self):
+    def close(self) -> None:
         self.stack.close()
 
 
@@ -301,34 +306,28 @@ class FlowPod(BasePod):
         return f'{cmd} {" ".join(self.cli_args)}'
 
     @staticmethod
-    def connect(first: 'BasePod', second: 'BasePod', first_socket_type: 'SocketType'):
+    def connect(first: 'BasePod', second: 'BasePod', first_socket_type: 'SocketType') -> None:
         """Connect two Pods
 
         :param first: the first BasePod
         :param second: the second BasePod
         :param first_socket_type: socket type of the first BasePod, availables are PUSH_BIND, PUSH_CONNECT, PUB_BIND
         """
-        if first_socket_type == SocketType.PUSH_BIND:
-            first.tail_args.socket_out = SocketType.PUSH_BIND
-            second.head_args.socket_in = SocketType.PULL_CONNECT
+        first.tail_args.socket_out = first_socket_type
+        second.head_args.socket_in = first.tail_args.socket_out.paired
 
+        if first_socket_type == SocketType.PUSH_BIND:
             first.tail_args.host_out = __default_host__
             second.head_args.host_in = _fill_in_host(bind_args=first.tail_args,
                                                      connect_args=second.head_args)
             second.head_args.port_in = first.tail_args.port_out
         elif first_socket_type == SocketType.PUSH_CONNECT:
-            first.tail_args.socket_out = SocketType.PUSH_CONNECT
-            second.head_args.socket_in = SocketType.PULL_BIND
-
             first.tail_args.host_out = _fill_in_host(connect_args=first.tail_args,
                                                      bind_args=second.head_args)
             second.head_args.host_in = __default_host__
             first.tail_args.port_out = second.head_args.port_in
         elif first_socket_type == SocketType.PUB_BIND:
-            first.tail_args.socket_out = SocketType.PUB_BIND
             first.tail_args.num_part += 1
-            second.head_args.socket_in = SocketType.SUB_CONNECT
-
             first.tail_args.host_out = __default_host__  # bind always get default 0.0.0.0
             second.head_args.host_in = _fill_in_host(bind_args=first.tail_args,
                                                      connect_args=second.head_args)  # the hostname of s_pod
@@ -338,7 +337,7 @@ class FlowPod(BasePod):
 
     def connect_to_tail_of(self, pod: 'BasePod'):
         """Eliminate the head node by connecting prev_args node directly to peas """
-        if self._args.replicas > 1 and self.is_head_router:
+        if self._args.parallel > 1 and self.is_head_router:
             # keep the port_in and socket_in of prev_args
             # only reset its output
             pod.tail_args = _copy_to_head_args(pod.tail_args, self._args.polling.is_push, as_router=False)
@@ -350,26 +349,26 @@ class FlowPod(BasePod):
             self.is_head_router = False
             self.deducted_head = pod.tail_args
         else:
-            raise ValueError('the current pod has no head router, deduct the head is confusing')
+            raise ValueError('the current pod has no head router, deducting the head is confusing')
 
     def connect_to_head_of(self, pod: 'BasePod'):
         """Eliminate the tail node by connecting next_args node directly to peas """
-        if self._args.replicas > 1 and self.is_tail_router:
+        if self._args.parallel > 1 and self.is_tail_router:
             # keep the port_out and socket_out of next_arts
             # only reset its input
             pod.head_args = _copy_to_tail_args(pod.head_args,
                                                as_router=False)
             # update peas to receive from it
             self.peas_args['peas'] = _set_peas_args(self._args, self.head_args, pod.head_args)
-            # remove the head node
+            # remove the tail node
             self.peas_args['tail'] = None
-            # head is no longer a router anymore
+            # tail is no longer a router anymore
             self.is_tail_router = False
             self.deducted_tail = pod.head_args
         else:
-            raise ValueError('the current pod has no tail router, deduct the tail is confusing')
+            raise ValueError('the current pod has no tail router, deducting the tail is confusing')
 
-    def start(self):
+    def start(self) -> 'FlowPod':
         if self._args.host == __default_host__:
             return super().start()
         else:
@@ -381,9 +380,9 @@ class FlowPod(BasePod):
             return self
 
 
-def _set_peas_args(args, head_args, tail_args):
+def _set_peas_args(args: Namespace, head_args: Namespace, tail_args: Namespace) -> List[Namespace]:
     result = []
-    for _ in range(args.replicas):
+    for _ in range(args.parallel):
         _args = copy.deepcopy(args)
         _args.port_in = head_args.port_out
         _args.port_out = tail_args.port_in
@@ -405,62 +404,58 @@ def _set_peas_args(args, head_args, tail_args):
     return result
 
 
-def _copy_to_head_args(args, is_push: bool, as_router: bool = True):
+def _copy_to_head_args(args: Namespace, is_push: bool, as_router: bool = True) -> Namespace:
     """Set the outgoing args of the head router
     """
 
     _head_args = copy.deepcopy(args)
     _head_args.port_ctrl = random_port()
     _head_args.port_out = random_port()
+    _head_args.uses = None
     if is_push:
         if args.scheduling == SchedulerType.ROUND_ROBIN:
             _head_args.socket_out = SocketType.PUSH_BIND
             if as_router:
-                _head_args.yaml_path = '_forward'
+                _head_args.uses = '_pass'
         elif args.scheduling == SchedulerType.LOAD_BALANCE:
             _head_args.socket_out = SocketType.ROUTER_BIND
             if as_router:
-                _head_args.yaml_path = '_route'
+                _head_args.uses = '_route'
     else:
         _head_args.socket_out = SocketType.PUB_BIND
-        _head_args.num_part = args.replicas
+        _head_args.num_part = args.parallel
         if as_router:
-            _head_args.yaml_path = '_forward'
+            _head_args.uses = '_pass'
 
     if as_router:
         _head_args.name = args.name or ''
         _head_args.role = PeaRoleType.HEAD
 
-    # head and tail never run in docker, reset their image to None
-    _head_args.image = None
     return _head_args
 
 
-def _copy_to_tail_args(args, as_router: bool = True):
+def _copy_to_tail_args(args: Namespace, as_router: bool = True) -> Namespace:
     """Set the incoming args of the tail router
     """
     _tail_args = copy.deepcopy(args)
     _tail_args.port_in = random_port()
     _tail_args.port_ctrl = random_port()
     _tail_args.socket_in = SocketType.PULL_BIND
+    _tail_args.uses = None
     if as_router:
-        _tail_args.yaml_path = args.reducing_yaml_path
+        _tail_args.uses = args.uses_reducing
         _tail_args.name = args.name or ''
         _tail_args.role = PeaRoleType.TAIL
 
-    # head and tail never run in docker, reset their image to None
-    _tail_args.image = None
     return _tail_args
 
 
-def _fill_in_host(bind_args, connect_args):
+def _fill_in_host(bind_args: Namespace, connect_args: Namespace) -> str:
     from sys import platform
 
     bind_local = (bind_args.host == '0.0.0.0')
-    bind_docker = (bind_args.image is not None and bind_args.image)
-    conn_tail = (connect_args.name is not None and connect_args.role == PeaRoleType.TAIL)
     conn_local = (connect_args.host == '0.0.0.0')
-    conn_docker = (connect_args.image is not None and connect_args.image)
+    conn_docker = (getattr(connect_args, 'uses', None) is not None and not valid_local_config_source(connect_args.uses))
     bind_conn_same_remote = not bind_local and not conn_local and (bind_args.host == connect_args.host)
     if platform == "linux" or platform == "linux2":
         local_host = '0.0.0.0'
@@ -483,7 +478,7 @@ def _fill_in_host(bind_args, connect_args):
 class GatewayPod(BasePod):
     """A :class:`BasePod` that holds a Gateway """
 
-    def start(self):
+    def start(self) -> 'GatewayPod':
         self.stack = ExitStack()
         for s in self.all_args:
             p = RESTGatewayPea(s) if getattr(s, 'rest_api', False) else GatewayPea(s)
